module Ogre

	# From Ogre's ExampleLoadingBar:
	#
	# Defines an example loading progress bar which you can use during
	# startup, level changes etc to display loading progress.
	#
	# Basically you just need to create an instance of this class, call start()
	# before loading and finish() afterwards. You may also need to stop areas of
	# your scene rendering in between since this method will call
	# RenderWindow::update() to update the display of the bar - we advise using
	# SceneManager's 'special case render queues' for this, see
	# SceneManager::addSpecialCaseRenderQueue for details.
	#	
	# This progress bar relies on you having the OgreCore.zip package already
	# added to a resource group called 'Bootstrap' - this provides the basic
	# resources required for the progress bar and will be loaded automatically.
	class ApplicationLoadingBar < ResourceGroupListenerProxy

		def start(window, num_groups_init = 1, num_groups_load = 1, initial_proportion = 0.7)
			@window = window
			@num_groups_init = num_groups_init
			@num_groups_load = num_groups_load
			@initial_proportion = initial_proportion

			ResourceGroupManager.instance.initialise_resource_group("Bootstrap")

			overlay_mgr = OverlayManager.instance
			@load_overlay = overlay_mgr.get_by_name("Core/LoadOverlay")

			unless @load_overlay
				raise "Cannot find loading overlay"
			end
			
			@load_overlay.show

			# Save links the loading text elements
			@loading_bar_element = overlay_mgr.get_overlay_element("Core/LoadPanel/Bar/Progress")
			@loading_comment_element = overlay_mgr.get_overlay_element("Core/LoadPanel/Comment")
			@loading_description_element = overlay_mgr.get_overlay_element("Core/LoadPanel/Description")

			bar = overlay_mgr.get_overlay_element("Core/LoadPanel/Bar")
			@progress_bar_max_size = bar.get_width
			@loading_bar_element.set_width(0)

			# See superclass
			register_as_listener
		end

		def finish
			@load_overlay.hide

			# See superclass
			unregister_as_listener
		end

		def resource_group_scripting_started(group_name, count)
			unless @num_groups_init > 0
				raise "You are initializing groups when you said you would not, failing before causing x / 0 error"
			end

			@progress_bar_inc = @progress_bar_max_size * @initial_proportion / (count * 1.0)
			@progress_bar_inc /= @num_groups_init

			@loading_description_element.set_caption("Parsing scripts...")
			@window.update
		end

		def script_parse_started(script_name)
			@loading_comment_element.set_caption(script_name)
			@window.update
		end

		def script_parse_ended(script_name)
			@loading_bar_element.set_width(@loading_bar_element.get_width + @progress_bar_inc)
			@window.update
		end

		def resource_group_scripting_ended(group_name)
		end

		def resource_group_load_started(group_name, count)
			puts "", "", "Resource group started #{group_name}", "", ""
			unless @num_groups_load > 0
				raise "You are initializing groups when you said you would not, failing before causing x / 0 error"
			end

			@progress_bar_inc = @progress_bar_max_size * (1 - @initial_proportion) / (count * 1.0)
			@progress_bar_inc /= @num_groups_load
			@loading_description_element.set_caption("Loading resources...")
			@window.update
		end

		def resource_load_started(resource)
			@loading_comment_element.set_caption(resource.get_name)
			@window.update
		end

		def resource_load_ended()
		end

		def world_geometry_stage_started(description)
			@loading_comment_element.set_caption(description)
			@window.update
		end

		def world_geometry_stage_ended
			@loading_bar_element.set_width(@loading_bar_element.get_width + @progress_bar_inc)
			@window.update
		end

		def resource_group_load_ended(name)
		end
	end
end
